"""
use this module to initialize the whole app

Note setuptools is not picking this up
probably because this module is directly off of src
"""
import qb_core.oproc_pkg.oproc
import qb_core.rats_pkg.rats
import qb_core.ocra_pkg.ocra


def initialize_app():
    qb_core.rats_pkg.rats.initialize_module()
    qb_core.oproc_pkg.oproc.initialize_module()
    qb_core.ocra_pkg.ocra.initialize_module()



import json

from qb_core.common.base_model import BaseModel


class OrderItem(BaseModel):
    def __init__(self):
        self.item_name = ""
        self.quantity = 0
        self.price = 0.0
        self.special_instruction = ""

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "item_name":
                self.item_name = v
            if k == "quantity":
                self.quantity = v
            if k == "price":
                self.price = v
            if k == "special_instruction":
                self.special_instruction = v

    def __str__(self):
        return f'{self.item_name=}, {self.quantity=}'


class OrderStatus(BaseModel):
    def __init__(self):
        self.time_order_created = ""
        self.time_driver_assigned = ""
        self.time_driver_picked_up_food = ""
        self.time_driver_delivered_food = ""

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "time_order_created":
                self.time_order_created = v
            if k == "time_driver_assigned":
                self.time_driver_assigned = v
            if k == "time_driver_picked_up_food":
                self.time_driver_picked_up_food = v
            if k == "time_driver_delivered_food":
                self.time_driver_delivered_food = v

    def __str__(self):
        return f'driver_assigned={self.time_driver_delivered_food != ""}, ' \
               f'driver_picked_up_food={self.time_driver_picked_up_food != ""}, ' \
               f'driver_delivered_food={self.time_driver_delivered_food != ""}'


class Order(BaseModel):
    def __init__(self):
        self.id = ""
        self.customer_email = ""
        self.total_price = 0.0
        self.items = []
        self.status = OrderStatus()
        self.status.time_order_created = "1/1/2022(TODO)"

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "total_price":
                self.total_price = v
            if k == "items":
                self.items = v
            if k == "id":
                self.id = v

    def __str__(self):
        return f'customer_email={self.customer_email}, ' \
               f'total_price={self.total_price}, ' \
               f'number_of_items={len(self.items)} ' \
               f'status=({str(self.status)})'


class BaseOrderEvent:
    def __init__(self, order):
        self.order = order

    def __repr__(self):
        return json.dumps(self.__dict__, default=str)

    def __str__(self):
        return f'{self.order=}'


class OrderAdded(BaseOrderEvent):
    pass


class OrderRemoved(BaseOrderEvent):
    pass



"""
this module has everything needed to represent Restaurant entity
this is similar to service object in spring boot

"""
# TODO figure how to wire this so that AWS flavor of the app can use DyanamoDB
from qb_core.rats_pkg.restaurant import _restaurant_persistence
from qb_core.rats_pkg.restaurant.restaurant_model import Restaurant


def does_restaurant_exist(name):
    r = _restaurant_persistence.read_restaurant(name)
    return r is not None


def create_restaurant(manager_name, manager_email, restaurant_name):
    r = Restaurant()
    r.init(manager_name=manager_name, manager_email=manager_email, name=restaurant_name)
    return r


def read_restaurant(restaurant_name):
    r = _restaurant_persistence.read_restaurant(restaurant_name)
    return r


def read_restaurant_by_manager_email(manager_email):
    r = _restaurant_persistence.read_restaurant_by_manager_email(manager_email)
    return r


def save_restaurant(restaurant):
    _restaurant_persistence.save_restaurant(restaurant)



from collections import defaultdict

ORDER_DICT = None
# example_order = {
#     "c1" : [Order1, Order2, Order3]
# }


def clear_all():
    global ORDER_DICT
    ORDER_DICT = defaultdict(list)


def read_order_count(customer_email):
    return len(ORDER_DICT[customer_email])


def read_orders(customer_email):
    return ORDER_DICT[customer_email]


def read_order(customer_email, order_id):
    return ORDER_DICT[customer_email]


def save_order(order):
    ORDER_DICT[order.customer_email].append(order)


clear_all()



from qb_core.rats_pkg.signin import _signin_persistence

ACTOR_RESTAURANT_MANAGER = "restaurant_manager"
ACTOR_CUSTOMER = "customer"
ACTOR_COOK = "cook"
ACTOR_DRIVER = "driver"


def signup(actor, email, password):
    return _signin_persistence.register_user(actor, email, password)


def signin(actor, email, password):
    return _signin_persistence.authenticate_user(actor, email, password)



from qb_core.oproc_pkg.order import _order_persistence


def read_open_orders():
    return _order_persistence.read_open_orders()


def save_order(order):
    print(f'OPROC order_entity.save_order called')
    _order_persistence.save_order(order)
    print(f'OPROC order_entity.save_order returns')



"""
rats.py
this follows interactor pattern.
implements both command executor and event listener

notes:
    this is the interactor Bob Martin talks about
    GOOD_TODO revisit this after you build core and infrastructure (or domain/infrastructure)
    I will implement two flavors of interactor: command executor and event listener
    for now both flavors will be here
"""
from qb_core.rats_pkg.cook import cook_entity
from qb_core.rats_pkg.cook.cook_model import Cook, CookAdded, CookRemoved
from qb_core.rats_pkg.signin import signin_service
from qb_core.rats_pkg.signup import signup_service
from qb_core.rats_pkg.restaurant import restaurant_entity
from qb_core.event_bus import event_bus_core, event_bus_events


def initialize_module():
    print(f'HEY HEY HEY initializing rats module HEY HEY HEY')
    event_bus_core.register_emitter(event_bus_events.EVENT_COOK_ADDED, None)
    event_bus_core.register_emitter(event_bus_events.EVENT_MENU_UPLOADED, None)
    print(f'*** REGISTERING listener None added for event restaurant_rated')
    event_bus_core.register_listener(event_bus_events.EVENT_RESTAURANT_RATED, None)


def signup_initiate(manager_name, manager_email, restaurant_name):
    # ensure manager_name, manager_email, restaurant_name is not empty
    exists = restaurant_entity.does_restaurant_exist(restaurant_name)
    if exists:
        return "error", "restaurant already signed up"

    restaurant = restaurant_entity.create_restaurant(manager_name, manager_email, restaurant_name)
    verification_code, state = signup_service.send_verification_email(manager_email, f'{restaurant_name=}')
    restaurant.status = state
    restaurant_entity.save_restaurant(restaurant)

    return "success", "restaurant signup initiated"


def signup_verify(manager_email, restaurant_name, code):
    # ensure manager_name, manager_email, restaurant_name is not empty
    restaurant = restaurant_entity.read_restaurant(restaurant_name)
    if restaurant is None:
        return "error", f'invalid {restaurant_name=}'

    result, message = signup_service.verify(manager_email, f'{restaurant_name=}', code)
    if result == "error":
        return result, message

    restaurant.status = result
    restaurant_entity.save_restaurant(restaurant)
    return "success", "restaurant signup verified"


# WARNING: do not expose this to clients e.g. don't expose this as API end points or even
# lambdas.
# provided this to help write low-level tests
def _signup_get_verification_code(manager_email, restaurant_name):
    return signup_service._get_verification_code(manager_email, f'{restaurant_name=}')


def signup_complete(restaurant_name, address, cuisine_type):
    # ensure manager_name, manager_email, restaurant_name is not empty
    restaurant = restaurant_entity.read_restaurant(restaurant_name)
    if restaurant is None:
        return "error", f'invalid {restaurant_name=}'

    restaurant.address = address
    restaurant.cuisine_type = cuisine_type
    restaurant.status = "signed-up"
    restaurant_entity.save_restaurant(restaurant)
    return "success", "restaurant signed-up"


def signup(manager_email, password):
    return signin_service.signup(signin_service.ACTOR_RESTAURANT_MANAGER, manager_email, password)


def signin(manager_email, password):
    return signin_service.signin(signin_service.ACTOR_RESTAURANT_MANAGER, manager_email, password)


# NICE_TODO Feature - Security verify security token
def add_cook(secure_token, cook_name, cook_email):
    restaurant = restaurant_entity.read_restaurant_by_manager_email(secure_token.email)
    cook = Cook()
    cook.init(name=cook_name, email=cook_email)
    cook_entity.add_cook(cook)
    restaurant.cook_dict[cook.name] = cook
    restaurant_entity.save_restaurant(restaurant)

    # this should be the last step so that we send this event exactly once
    # when all the steps above have completed
    event_bus_core.emit_event(event_bus_events.EVENT_COOK_ADDED, CookAdded(cook))
    return cook


def remove_cook(secure_token, cook):
    restaurant = restaurant_entity.read_restaurant_by_manager_email(secure_token.email)
    restaurant.cook_dict.pop(cook.email, None)
    restaurant_entity.save_restaurant(restaurant)
    cook_entity.remove_cook(cook)

    # this should be the last step so that we send this event exactly once
    # when all the steps above have completed
    event_bus_core.emit_event("cook_removed", CookRemoved(cook))


def upload_menu():
    pass


def view_order_status():
    pass


def restaurant_rated():
    pass







from collections import defaultdict
from random import randrange

from qb_core.rats_pkg.signin.secure_context import SecureToken

SIGNIN_DICT = None
SIGNIN_SESSION = None


def clear_all():
    global SIGNIN_DICT
    SIGNIN_DICT = defaultdict(dict)
    global SIGNIN_SESSION
    SIGNIN_SESSION = defaultdict(list)


def authenticate_user(actor, email, password):
    secure_token = None
    if SIGNIN_DICT[actor].get(email) == password:
        secure_token = SecureToken()
        secure_token.email = email
        secure_token.access_token = generate_token()
        secure_token.refresh_token = generate_token()
        # NICE_TODO Feature - Security set this to a proper time 5 minutes
        secure_token.access_token_expires_at = 5 * 60
        # NICE_TODO Feature - Security set this to a proper time 30 minutes
        secure_token.refresh_token_expires_at = 30 * 60
        # only one session per user/email
        SIGNIN_SESSION[email] = secure_token
    return secure_token


def register_user(actor, email, password):
    # print(f'{actor=}, {email=}')
    SIGNIN_DICT[actor][email] = password
    # print('done adding stuff')
    return "success", f'signed-up {actor=}, {email=}'


# hey, lets not get into the habit of printing sensitive info
# so turn this off
# WARNING never turn this ON
# def print_all():
#     print(f'{SIGNIN_DICT=}')


def generate_token():
    r1 = randrange(10)
    r2 = randrange(10)
    r3 = randrange(10)
    r4 = randrange(10)
    r5 = randrange(10)
    r6 = randrange(10)
    r7 = randrange(10)
    r8 = randrange(10)
    r9 = randrange(10)
    r = str(r1) + str(r2) + str(r3) + str(r4) + str(r5) + str(r6) + str(r7) + str(r8) + str(r9)
    return r


clear_all()



from random import randrange

from qb_core.rats_pkg.signup import _signup_persistence


def generate_code():
    r1 = randrange(10)
    r2 = randrange(10)
    r3 = randrange(10)
    r4 = randrange(10)
    r5 = randrange(10)
    r6 = randrange(10)
    r = str(r1) + str(r2) + str(r3) + str(r4) + str(r5) + str(r6)
    return r


def send_verification_email(email, tag):
    code = generate_code()
    # NICE_TODO Feature - Security send email
    print(f'TODO send email to {email}')
    print(f'TODO persist state {code=}, {tag=}')
    key = make_key(email, tag)
    _signup_persistence.save_code(key, code)
    state = "pending_verification"
    return code, state


def verify(email, tag, code):
    key = make_key(email, tag)
    if _signup_persistence.read_code(key) == code:
        state = "verified"
    else:
        state = "error"
    return state, "invalid key"


def make_key(email, tag):
    return f'{email=}#{tag=}'


def _get_verification_code(email, tag):
    key = make_key(email, tag)
    return _signup_persistence.read_code(key)
